# Step 1: in the doc folder run
#         make clean
#         DUNE_LOGMODULES=8 make -ij8
# Step 2: in this folder run
#         python generate.py

import os, shutil, sys
from dune.common.module import getDunePyDir
from importlib.metadata import version
femVersion = version("dune.fem")
def main():
    if len(sys.argv)>1:
        path = sys.argv[1:]
    else:
        path = ["doc"]
    docPath = os.path.join("..","..",*path)
    print(f"using modules files from {docPath}")
    if not os.path.exists(docPath):
        raise RuntimeError(f"folder {docPath} does not exist")

    if os.path.exists(femVersion):
        shutil.rmtree(femVersion)
    os.makedirs(femVersion)

    dunepy = getDunePyDir()
    names = set()
    for file in os.listdir(docPath):
        filename = os.path.join(docPath,os.fsdecode(file))
        if filename.endswith(".modules"):
            with open(filename,"r") as f:
                names.update(f.readlines())
            shutil.copy(filename, femVersion)

    for name in names:
        name = name.strip()
        print(name)
        src = os.path.join(dunepy,"python","dune","generated",name+".cc")
        try:
            shutil.copy(src, femVersion)
        except FileNotFoundError:
            print(f"Error: can't copy {src}",flush=True)
            pass

if __name__ == "__main__":
    main()



from . import geometry
from . import transformers
from . import boundary
from .model2ufl import model2ufl
from .plot import plotSolution

# a utility method to construct a suitable grid
def GridView(domain, omega=None, *,
                filterTolerance=0,
                factor=0, refinements=0,
                adaptLevels=0):
    import dune.ufl
    from dune.fem.function import gridFunction
    from dune.fem import adapt, mark, markNeighbors
    from dune.fem.view import adaptiveLeafGridView, filteredGridView
    from dune.grid import cartesianDomain, Marker
    from dune.alugrid import aluConformGrid as GridView
    import ufl
    if hasattr(domain,"indexSet"):
        return domain
    if type(domain) == list:
        domain = cartesianDomain(*domain)
    gridView = adaptiveLeafGridView( GridView(domain) )
    dims = (gridView.dimGrid, gridView.dimWorld)
    x = ufl.SpatialCoordinate(dune.ufl.domain(dims))

    if omega and factor>0 and refinements>0:
        startEps = omega.epsilon.value * factor
        region = gridFunction(
            omega.phi(x) * (1 - omega.phi(x)), gridView=gridView
        )  # interface
        for j in range(1, refinements + 1):
            omega.epsilon.value = startEps * j / refinements
            mark(region, 0.00247262315663, maxLevel=refinements)  # 1 epsilon
            adapt(gridView.hierarchicalGrid)
        omega.epsilon.value = startEps/factor * 0.5**(refinements/dims[0])
    if adaptLevels>0:
        sd = gridFunction(omega(x), gridView=gridView)
        bary = dims[0] * [1.0 / (dims[0]+1)]
        lowerTol=-0.1
        upperTol=0.1
        for _ in range(adaptLevels):
            def markStrategy(e):
                v = sd(e, bary)
                return (
                    Marker.refine
                    if v > lowerTol and v < upperTol
                    else Marker.keep
                )
            gridView.hierarchicalGrid.adapt(markStrategy)
            lowerTol /= 2
            upperTol /= 2

    if filterTolerance<0:
        return gridView
    if omega and filterTolerance>0:
        sd = gridFunction(omega(x), gridView=gridView)
        bary = dims[0] * [1.0 / (dims[0]+1)]
        filter = lambda e: 1 if sd(e, bary) < filterTolerance else 2
    else:
        filter = lambda e: 1
    return filteredGridView(
               gridView, filter, domainId=1, useFilteredIndexSet=True)

def solve(uflModel, domain):
    # we use a Lagrange space
    from dune.fem.space import lagrange
    # and the build-in solver
    from dune.fem.scheme import galerkin

    gridView = GridView(domain)
    space = lagrange(gridView, dimRange=1)
    scheme = galerkin(uflModel, space=space,
                      solver=("suitesparse","umfpack"))
    uh = space.interpolate(0,name="ddm")
    scheme.solve(target=uh)
    return uh



from dune.ufl import Constant
from ufl import as_vector, conditional, dot, grad, outer, zero

from ddfem.boundary import BndFlux_c, BndFlux_v, BndValue


def crModel():
    class PotentialModel:
        dimRange = 1
        outFactor_i = Constant(1, "outFactor")

        def F_v(t, x, U, DU):
            return DU

        def S_e(t, x, U, DU):
            return as_vector([-1])

        def valD(t, x, U):
            return zero(1)

        boundary = {
            "full": BndValue(valD),
        }

    class ChemModel:
        dimRange = 3

        dt = Constant(0.05, "dt")
        diff = Constant(1e-3, "diff")  # this is about the boundary for stability
        # outFactor_i = diff

        P1 = as_vector([-0.25, -0.25])  # midpoint of first source (close to boundary)
        P2 = as_vector([0.25, 0.25])  # midpoint of second source (close to boundary)

        u_h_n = None
        dpsi_h = None

        def setup(u_h, DDMPotentialModel, psi_h):
            ChemModel.u_h_n = u_h.copy(name="u_h_n")
            ChemModel.u_h_n.interpolate(zero(ChemModel.dimRange))

            ChemModel.dpsi_h = lambda x: DDMPotentialModel.sigma(
                0, x, psi_h, grad(psi_h)
            )

        def f1(x):
            q = x - ChemModel.P1
            return conditional(dot(q, q) < 0.04**2, 5, 0)

        def f2(x):
            q = x - ChemModel.P2
            return conditional(dot(q, q) < 0.04**2, 5, 0)

        def f(t, x):
            return conditional(
                t < 10,
                as_vector([ChemModel.f1(x), ChemModel.f2(x), 0]),
                as_vector([0, 0, 0]),
            )

        def r(U):
            return 10 * as_vector([U[0] * U[1], U[0] * U[1], -2 * U[0] * U[1]])

        def F_v(t, x, U, DU):
            return ChemModel.diff * DU

        def velocity(x):
            return as_vector([-ChemModel.dpsi_h(x)[0, 1], ChemModel.dpsi_h(x)[0, 0]])

        def F_c(t, x, U):
            return outer(U, ChemModel.velocity(x))

        def S_i(t, x, U, DU):
            return -(U - ChemModel.u_h_n) / ChemModel.dt

        def S_e(t, x, U, DU):
            return ChemModel.f(t, x) - ChemModel.r(ChemModel.u_h_n)

        # boundary = {"full": BndValue(lambda t, x, U: zero(3))}

        boundary = {
            "full": [
                BndFlux_c(lambda t, x, U, n: zero(ChemModel.dimRange)),
                BndFlux_v(lambda t, x, U, DU, n: zero(ChemModel.dimRange)),
            ]
        }

    return PotentialModel, ChemModel



from dune.ufl import Constant
from ufl import as_vector, conditional, div, dot, exp, grad, inner, sqrt, outer, zero
import ufl
from ddfem.boundary import AFluxBC, DFluxBC, ValueBC

def CModel(inverted):
    withAdv = False

    class Model:
        dimRange = 1
        stabFactor = Constant(1)

        def initial(x):
            return 2

        def exact(t, x):
            return as_vector((2+x[0]-x[0],))   # better way of avoiding 'Cannot determine geometric dimension from expression'

        def K(U, DU):
            return 1e-2

        def F_v(t, x, U, DU):
            return Model.K(U, DU) * DU

        if withAdv:
            def F_c(t, x, U):
                return outer(U, as_vector([-2,0.5]) )
  
        if withAdv:
            bndFlux = [ AFluxBC( lambda t,x,U,n: Model.F_c(t,x,U)*n ),
                        DFluxBC( lambda t,x,U,DU,n: Model.F_v(t,x,U,DU)*n ) ]
        else:
            # this works: bndFlux = DFluxBC( lambda t,x,U,DU,n: zero(U.ufl_shape)
            bndFlux = DFluxBC( lambda t,x,U,DU,n: Model.F_v(t,x,U,DU)*n )
        boundary = {
            # "full": ValueBC( lambda t,x,U: as_vector([2.]) ),
            "full": bndFlux
            # "sides": ValueBC( lambda t,x,U: as_vector([2.]) ),
            # "ends": [AFluxBC( lambda t,x,U,n: Model.F_c(t,x,U)*n ),
            #          DFluxBC( lambda t,x,U,DU,n: Model.F_v(t,x,U,DU)*n )]
        }

        if inverted:
            boundary[range(1,5)] = ValueBC( lambda t,x,U: as_vector([2.]) )

    return Model



from ufl import as_vector, cos, dot, max_value, min_value, sin

from .helpers import ufl_clamp, ufl_cross, ufl_length, ufl_sign
from .primitive_base import ORIGIN, SDF


class Pie(SDF):
    def __init__(self, radius, angle, *args, **kwargs):
        self.radius = radius
        self.angle = angle  # angle of cicle (not opening)

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Pie({self.radius}, {self.angle}, {self._repr_core()})"

    def sdf(self, x):
        # Note: Ignores z
        x0_abs = abs(x[0])
        coords = as_vector([x0_abs, x[1]])

        circle_dist = ufl_length(coords) - self.radius

        trig = as_vector([sin(self.angle / 2), cos(self.angle / 2)])

        # projection of coords on to trig, clamped to within circle
        proj = ufl_clamp(dot(coords, trig), 0, self.radius) * trig
        rejc = coords - proj
        edge_dist = ufl_length(rejc) * ufl_sign(ufl_cross(coords, trig))

        return max_value(circle_dist, edge_dist)



from ufl import grad, outer

from .transformer_base import transformer

"""
Diffusion:
Paper with sigma = grad(phi u) - grad(phi)g
phi Asigma.(grad(phi v) + grad(phi)v) + 1/4 (Agrad(phi),grad(phi))(u-g)v
= phi Asigma.(phi grad(v) + 2grad(phi)v) + 1/4 (Agrad(phi),grad(phi))(u-g)v
= phi^2 Asigma.grad(v) + 2phi Asigma.grad(phi)v + 1/4 (Agrad(phi),grad(phi))(u-g)v

orig.F_v(u,Du) = A Du
orig.S_i(u,Du) = -div(ADg)

mix0.F_v = phi^2 orig.F_v(u,sigma)
mix0.S_i = phi^2 orig.S_i(u,sigma) - 2phi orig.F_v(u,sigma).grad(phi) - 1/4 F_v(u,grad(phi)(u-g)).grad(phi)

model = mix0.F_v(u,Du).grad(v) - mix0.S_i(u,Du)v
      = phi^2 Asigma.grad(v) + phi^2 div(ADg)v + 2phi Asigma.grad(phi)v + 1/4 Agrad(phi)(u-g).grad(phi)
      = phi Asigma.grad(phi v) - phi Asigma.grad(phi)v + phi^2 div(ADg)v + 2phi Asigma.grad(phi)v
        + 1/4 Agrad(phi).grad(phi)(u-g)
      = phi Asigma.(grad(phi v) + grad(phi)v) + phi^2 div(ADg)v + 1/4 Agrad(phi).grad(phi))(u-g)

Advection:
orig.F_c = bu
orig.S_e = div(bg)

mix0.F_c(u) = phi^2 orig.F_c(u)
            = phi^2 bu
mix0.S_e(u) = phi^2 orig.S_e(u) - phi orig.F_c(u+g).grad(phi)
            = phi^2 div(bg) + phi (u+g) b.grad(phi)

model = - mix0.F_c(u).grad(v) - mix0.S_e(u)v
      = - phi^2 u b.grad(v) - phi^2 div(bg)v - phi (u+g) b.grad(phi)v
      = ( div(phi u phi b) - phi (u+g) b.grad(phi) - phi^2 div(bg) )v
      = ( phi b.grad(phi u) + phi u div(phi b) - phi (u+g) b.grad(phi) - phi^2 div(bg) )v
      = ( phi b.(grad(phi u)-g grad(phi))
          - phi u b.grad(phi) + phi^2 u div(b) + phi u b.grad(phi)
          - phi^2 b.grad(g) - phi^2 g div(b) )v
      = ( phi b.(phi sigma) + phi^2 div(b)(u-g) - phi^2 b.grad(g) )v
      = phi^2 ( b.(sigma - grad(g)) + div(b)(u-g) ) v

paper = phi b . sigma v
      = phi b . [grad(phi u) - g.grad(phi)] v
"""


@transformer
def Mix0DDM(Model):

    class DDModel(Model):
        def sigma(t, x, U, DU=None):
            if not DU:
                DU = grad(U)
            sigma = DU + outer(
                DDModel.BT.jumpV(t, x, U), grad(DDModel.phi(x))
            ) / DDModel.phi(x)
            return sigma

        def S_e_source(t, x, U, DU):
            return DDModel.phi(x) ** 2 * Model.S_e(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_e_convection(t, x, U, DU):
            # unsure if using Model.F_c is the right choice here - works in the linear case.
            convec = (
                DDModel.phi(x)
                * Model.F_c(t, x, 2 * U - DDModel.BT.jumpV(t, x, U))
                * grad(DDModel.phi(x))
            )

            convec -= DDModel.phi(x) * DDModel.BT.BndFlux_cExt(t, x, U)
            return convec

        def S_outside(t, x, U, DU):
            return -(
                DDModel.BT.jumpV(t, x, U)
                * (1 - DDModel.phi(x)) ** 2
                / (DDModel.epsilon**2)
            )

        def S_i_source(t, x, U, DU):
            return DDModel.phi(x) ** 2 * Model.S_i(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_i_diffusion(t, x, U, DU):
            Fv = Model.F_v(t, x, U, DDModel.sigma(t, x, U, DU))
            diffusion = 2 * DDModel.phi(x) * Fv * grad(DDModel.phi(x))

            gp = grad(DDModel.phi(x))
            uogp = outer((DDModel.BT.jumpV(t, x, U)), gp)
            # use of F_v is probably not correct...
            diffusion += (
                # Model.F_v_lin_mult(t, x, U, Mix0DDM.sigma(t,x, U, DU), uogp)
                Model.F_v(t, x, U, uogp)
                * gp
                / 4
            )

            diffusion += DDModel.phi(x) * DDModel.BT.jumpFv(t, x, U, DU, Fv)
            return -diffusion

        def F_c(t, x, U):
            return DDModel.phi(x) ** 2 * Model.F_c(t, x, U)

        def F_v(t, x, U, DU):
            return DDModel.phi(x) ** 2 * Model.F_v(t, x, U, DDModel.sigma(t, x, U, DU))

    return DDModel



from ufl import as_vector, conditional, dot, grad, outer, sqrt, zero

from .transformer_base import transformer

"""
Paper: using x^- = 1/2(x-|x|), x^+ = 1/2(x+|x|) so that x = x^+ + x^-, i.e., x^+ - x = - x^-
the advection b.grad(u)v is replaced with
phi b.grad(u) v + [b.grad(phi)]^+ (u-g)v
= div(phi bu)v - div(phi b) uv + [b.grad(phi)]^+ (u-g)v
= - phi u b.grad(v) - phi div(b) uv - b.grad(phi) uv + [b.grad(phi)]^+ uv - [b.grad(phi)]^+ gv
= - phi u b.grad(v) - phi div(b) uv + ([b.grad(phi)]^+ - b.grad(phi)) uv - [b.grad(phi)]^+ gv
= - phi u b.grad(v) - phi div(b) uv - [b.grad(phi)]^- uv - [b.grad(phi)]^+ gv

orig.F_c = bu
orig.S_e = div(bg)

ddm2.F_c(u) = phi orig.F_c(u)
            = phi bu
ddm2.S_e(u) = phi orig.S_e(u) + [b.grad(phi)]^- u
            = phi div(bg) + [b.grad(phi)]^- u + [b.grad(phi)]^+ gv

model = - ddm2.F_c(u).grad(v) - ddm2.S_e(u)v
      = - phi u b.grad(v) - phi div(bg)v - [b.grad(phi)]^- uv - [b.grad(phi)]^+ gv
      = div(phi u b)v - phi div(b)gv - phi b.grad(g) v - [b.grad(phi)]^- uv - [b.grad(phi)]^+ gv
      = phi b.grad(u)v + div(phi b)uv - phi div(b)gv - phi b.grad(g) v - [b.grad(phi)]^- uv - [b.grad(phi)]^+ gv
      = phi b.grad(u)v + phi div(b) uv + b.grad(phi)uv - [b.grad(phi)]^- uv - phi div(b)gv - phi b.grad(g) v - [b.grad(phi)]^+ gv
      = phi b.(grad(u) - grad(g)) v + phi div(b) (u-g)v + b.grad(phi)uv - [b.grad(phi)]^- uv - [b.grad(phi)]^+ gv
      = phi b.(grad(u) - grad(g)) v + phi div(b) (u-g)v + [b.grad(phi)]^+ (u-g)v
"""


@transformer
def NSDDM(Model):
    class DDModel(Model):
        def S_e_source(t, x, U, DU):
            return DDModel.phi(x) * Model.S_e(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_e_convection(t, x, U, DU):
            if DDModel.BT.BndValueExt is not None:
                direction = Model.F_c_lin_mult(t, x, U, -grad(DDModel.phi(x)))  # fc * n
                flux_u = Model.F_c(t, x, U) * -grad(DDModel.phi(x))
                flux_g = Model.F_c(t, x, DDModel.BT.BndValueExt(t, x, U)) * -grad(
                    DDModel.phi(x)
                )
                convec = []
                for i in range(U.ufl_shape[0]):
                    convec.append(
                        conditional(direction[i, i] > 0, flux_u[i], flux_g[i])
                    )
                convec = -as_vector(convec)
            else:
                convec = zero(U.ufl_shape)

            convec -= DDModel.BT.BndFlux_cExt(t, x, U)
            return convec

        def S_outside(t, x, U, DU):
            return -(
                DDModel.BT.jumpV(t, x, U) * (1 - DDModel.phi(x)) / (DDModel.epsilon**2)
            )

        def S_i_source(t, x, U, DU):
            return DDModel.phi(x) * Model.S_i(t, x, U, DDModel.sigma(t, x, U, DU))

        def S_i_diffusion(t, x, U, DU):
            beta = 6 * (1 - DDModel.phi(x)) / DDModel.epsilon
            diffusion = beta * (
                sqrt(dot(grad(DDModel.phi(x)), grad(DDModel.phi(x))))
                * DDModel.BT.jumpV(t, x, U)
            )
            Fv = Model.F_v(t, x, U, DDModel.sigma(t, x, U, DU))
            diffusion += Fv * grad(DDModel.phi(x))
            diffusion += DDModel.BT.jumpFv(t, x, U, DU, Fv)
            return -diffusion

        def F_c(t, x, U):
            return DDModel.phi(x) * Model.F_c(t, x, U)

        def F_v(t, x, U, DU):
            diffusion = DDModel.phi(x) * Model.F_v(t, x, U, DDModel.sigma(t, x, U, DU))

            out = outer(DDModel.BT.jumpV(t, x, U), grad(DDModel.phi(x)))
            diffusion += Model.F_v_lin_mult(t, x, U, DDModel.sigma(t, x, U, DU), out)
            return diffusion

    return DDModel



from dune.ufl import Constant
from ufl import Identity, as_vector, div, grad, sym, tr, zero

from ddfem.boundary import BndFlux_v, BndValue


def leModel():
    class Model:
        dimRange = 2
        lamb = Constant(0.1, name="Lame_1")
        mu = Constant(1, name="Lame_2")
        rho = Constant(1 / 1000, name="density")
        g = Constant(9.8, name="gravity")

        outFactor_i = Constant(1, "outFactor")

        vd = Constant([0, 0], "stationary")
        vn = Constant([0, 0], "traction")

        I = Identity(dimRange)

        def sigma(U, DU):
            return Model.lamb * tr(sym(DU)) * Model.I + 2 * Model.mu * sym(DU)

        def F_v(t, x, U, DU):
            return Model.sigma(U, DU)

        def S_i(t, x, U, DU):
            return as_vector([0.0, -Model.rho * Model.g])

        valD = lambda t, x, U: Model.vd

        valN = lambda t, x, U, DU, n: Model.vn

        # boundary = {
        #     "left": BndValue(valD),
        #     "other": BndFlux_v(valN),
        # }
        boundary = {
            "left": BndValue(valD),
            "right": BndFlux_v(valN),
            "bulk": BndFlux_v(valN),
        }

    return Model



from ufl import as_vector, conditional, cos, sin

from .helpers import ufl_length
from .primitive_base import ORIGIN, SDF


class Arc(SDF):
    """SDF for an arc:

    Provides the signed distance function for an arc given the
    radius, angle of opening, the width, and the center (which defaults to the origin).
    """

    def __init__(self, radius, angle, width, center=ORIGIN, *args, **kwargs):
        self.radius = radius
        self.angle = angle  # angle of opening
        self.width = width

        if isinstance(center, (list, tuple)):
            center = as_vector(center)
        self.center = center

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Arc({self.radius}, {self.angle}, {self.width}, {self.center}, {self._repr_core()})"

    def sdf(self, x):
        # Note: Ignores z
        y0_abs = abs(x[1])
        coords = as_vector([x[0], y0_abs])

        center_radius = self.radius - self.width / 2

        distance = ufl_length(coords) - center_radius

        edge_point = center_radius * as_vector([cos(self.angle), sin(self.angle)])

        left_coords = coords - edge_point

        sign_dist = conditional(
            (sin(self.angle) * x[0]) > (cos(self.angle) * y0_abs),
            ufl_length(left_coords),
            abs(distance),
        )

        sign_dist -= self.width / 2
        return sign_dist



import time

import dune.ufl
from dune.fem.function import gridFunction
from dune.fem import adapt, mark, markNeighbors
from dune.fem.view import adaptiveLeafGridView, filteredGridView
from dune.grid import cartesianDomain, Marker
from dune.alugrid import aluConformGrid as GridView
from dune.fem.utility import FemThreadPoolExecutor
import ufl

from .domain import Domain
from .. import GridView

class DomainDune(Domain):
    def __init__(self, omega, domain, *,
                 filterTolerance=0,
                 factor=0, refinements=0,
                 adaptLevels=0):
        super().__init__(omega)
        if not hasattr(omega.epsilon,"value"):
            omega.epsilon = dune.ufl.Constant(omega.epsilon,name="epsilon")
        self.gridView = GridView(domain, omega,
                              filterTolerance=filterTolerance,
                              factor=factor,refinements=refinements,
                              adaptLevels=adaptLevels)
        dims = (self.gridView.dimGrid, self.gridView.dimWorld)
        self.x = ufl.SpatialCoordinate(dune.ufl.domain(dims))
        self.fullSDF = self.gridFunction(self.omega(self.x), name="full-sdf")
        self._phi = None
        self._bndProj = None
        self._extProj = None
        self._bndProjSDFs = {}
        self._chi = None

        """
        print("Setting up projections",flush=True)
        with FemThreadPoolExecutor(max_workers=8) as executor:
            executor.submit( self.boundary_projection, self.x )
            executor.submit( self.external_projection, self.x )
            executor.submit( self.phi, self.x )
            executor.submit( self.chi, self.x )
            for child in self.omega:
                executor.submit( self.generate_projSDF, child )
        print("done",flush=True)
        """

    def gridFunction(self, expr, name):
        start_ = time.time()
        gf = gridFunction(expr, name=name, gridView=self.gridView)
        # print(f"{name} setup: {time.time() - start_}")
        return gf

    def phi(self, x):
        if self._phi is None:
            p = super().phi(self.x)
            self._phi = self.gridFunction(p, "phidomain")

        return self._phi

    def chi(self, x):
        return super().chi(self.x)
        if self._chi is None:
            p = super().chi(self.x)
            self._chi = self.gridFunction(p, "chidomain")

        return self._chi

    def boundary_projection(self, x):
        if self._bndProj is None:
            p = super().boundary_projection(self.x)
            self._bndProj = self.gridFunction(p, "bndproj")

        return self._bndProj

    def external_projection(self, x):
        if self._extProj is None:
            p = super().external_projection(self.x)
            self._extProj = self.gridFunction(p, "extproj")

        return self._extProj

    def generate_projSDF(self, sdf):
        projSDF = self._bndProjSDFs.get(sdf.name)
        if projSDF is None:
            projSDF = super().generate_projSDF(sdf)
            gf = self.gridFunction(projSDF(self.x), f"sdfproj{sdf.name}")
            self._bndProjSDFs[sdf.name] = lambda x: gf
            projSDF = self._bndProjSDFs[sdf.name]
        return projSDF

    def _adapt(self, level, lowerTol=-0.1, upperTol=0.1):
        for _ in range(level):
            def mark(e):
                v = self.fullSDF(e, self.gridView.dimension * [1.0 / 3.0])
                return (
                    Marker.refine
                    if v > lowerTol and v < upperTol
                    else Marker.keep
                )
            self.gridView.hierarchicalGrid.adapt(mark)
            lowerTol /= 2
            upperTol /= 2



from functools import reduce

from ufl import conditional, eq, grad, min_value

from ..boundary import BndFlux_c, BndFlux_v, BndValue, boundary_validation
from ..geometry.domain import Domain
from ..geometry.primitive_base import SDF
from .transformer_base import pretransformer


def Fitted(OriginalModel, domainDescription):

    if isinstance(domainDescription, Domain):
        domain = domainDescription
    else:
        domain = Domain(domainDescription)

    condition = lambda k: isinstance(k, str) or isinstance(k, SDF)

    class Fitted(OriginalModel):
        def sigma(t, x, U, DU=None):
            if DU:
                return DU
            return grad(U)

        diffuse = {k: v for k, v in OriginalModel.boundary.items() if condition(k)}
        boundary = {k: v for k, v in OriginalModel.boundary.items() if not condition(k)}

        bndSDFs = {k: domain.bndSDFs(k) for k in diffuse.keys()}

        boundary_flux_cs, boundary_flux_vs, boundary_values = boundary_validation(
            OriginalModel, override_boundary_dict=diffuse
        )

        def make_boundary_function(key, mv, bndSDFs=bndSDFs):
            sdf = bndSDFs[key]
            closest_sdf = lambda x: reduce(
                min_value,
                ([abs(v(x)) for b, v in bndSDFs.items()]),
            )

            boundary_map = lambda x: conditional(eq(closest_sdf(x), abs(sdf(x))), 1, 0)

            if isinstance(mv, BndFlux_v):
                return BndFlux_v(
                    lambda t, x, u, DU, n: boundary_map(x) * mv(t, x, u, DU, n),
                )

            elif isinstance(mv, BndFlux_c):
                return BndFlux_c(
                    lambda t, x, u, n: boundary_map(x) * mv(t, x, u, n),
                )

        def make_boundary_conditional(key, bndSDFs=bndSDFs, tol=1e-2):
            sdf = bndSDFs[key]
            return lambda x: abs(sdf(x)) < tol

        for bc_key, bc_value in boundary_values.items():
            boundary[make_boundary_conditional(bc_key)] = bc_value

        for bc_key in boundary_flux_cs.keys() | boundary_flux_vs.keys():
            if bc_key in boundary_flux_cs and bc_key in boundary_flux_vs:
                af = make_boundary_function(bc_key, boundary_flux_cs[bc_key])
                df = make_boundary_function(bc_key, boundary_flux_vs[bc_key])
                boundary[make_boundary_conditional(bc_key)] = [af, df]

            elif bc_key in boundary_flux_cs and bc_key not in boundary_flux_vs:
                af = make_boundary_function(bc_key, boundary_flux_cs[bc_key])
                boundary[make_boundary_conditional(bc_key)] = af

            elif bc_key not in boundary_flux_cs and bc_key in boundary_flux_vs:
                df = make_boundary_function(bc_key, boundary_flux_vs[bc_key])
                boundary[make_boundary_conditional(bc_key)] = df
            else:
                raise ValueError()

    return Fitted



from ufl import conditional, dot, grad, sqrt

from .primitive_base import SDF


class Domain:
    def __init__(self, omega):
        self.omega = omega

    def __call__(self, x):
        return self.omega(x)

    def phi(self, x):
        tol = 1e-10
        return (1 - tol) * self.omega.phi(x) + tol

    def chi(self, x):
        return self.omega.chi(x)

    def scaled_normal(self, x):
        return -grad(self.phi(x))

    def surface_delta(self, x):
        return sqrt(dot(self.scaled_normal(x), self.scaled_normal(x)))

    def normal(self, x):
        tol = 1e-10
        sd = conditional(self.surface_delta(x) > tol, self.surface_delta(x), tol)
        return self.scaled_normal(x) / sd  # grad(self.omega(x))

    def boundary_projection(self, x):
        return x + self.omega.projection(x)

    def external_projection(self, x):
        return x + self.omega.projection(x) * (1 - self.chi(x))

    def bndSDFs(self, SDFname):
        if isinstance(SDFname, SDF):
            SDFname = SDFname.name

        sdf = self.omega.search(SDFname)
        if sdf is None:
            raise ValueError(f"No SDF with name {SDFname}")
        return sdf

    def bndProjSDFs(self, SDFname):
        sdf = self.bndSDFs(SDFname)
        return self.generate_projSDF(sdf)

    def generate_projSDF(self, sdf):
        w = lambda x: sdf.phi(x) * (1 - sdf.phi(x))
        return lambda x: w(self.boundary_projection(x))



from dune.ufl import Constant
from ufl import grad, inner

from ddfem.boundary import BndValue


def pModel(power):
    class Model:
        dimRange = 1
        ep = Constant(1e-5, name="ep")
        f = Constant([1], name="f")
        g = Constant([0], name="g")
        p = Constant(power, name="p")  # p for p-Laplacian
        assert p.value > 1 and p.value < 2**16
        outFactor_i = Constant(1, "outFactor")

        def K(u):
            return (Model.ep**2 + inner(grad(u), grad(u))) ** ((Model.p - 2) / 2)

        def F_v(t, x, U, DU):
            return Model.K(U) * DU

        def S_i(t, x, U, DU):
            return Model.f - U

        bc = BndValue(g)
        boundary = {1: bc, 2: bc, 3: bc, 4: bc, "full": bc}

    return Model



from ufl import (
    FacetNormal,
    SpatialCoordinate,
    TestFunction,
    TrialFunction,
    as_vector,
    ds,
    dx,
    grad,
    inner,
)

from .boundary import boundary_validation


def boundaries_ufl(Model, space, t):
    boundary_flux_cs, boundary_flux_vs, boundary_values = boundary_validation(Model)

    u = TrialFunction(space)
    n = FacetNormal(space.cell())
    x = SpatialCoordinate(space.cell())

    boundary_flux_cs = {
        (k(x) if callable(k) else k): f(t, x, u, n) for k, f in boundary_flux_cs.items()
    }
    boundary_flux_vs = {
        (k(x) if callable(k) else k): f(t, x, u, grad(u), n)
        for k, f in boundary_flux_vs.items()
    }
    boundary_values = {
        (k(x) if callable(k) else k): f(t, x, u) for k, f in boundary_values.items()
    }
    hasBoundaryValue = {k: True for k in boundary_values.keys()}

    return (
        boundary_flux_cs,
        boundary_flux_vs,
        boundary_values,
        hasBoundaryValue,
    )


class DirichletBC:
    def __init__(self, space, value, domain=None):
        self.space = space
        self.value = value
        self.domain = domain

    def __str__(self):
        return str(self.value) + str(self.domain)


def model_ufl(Model, space, initialTime=0, DirichletBC=DirichletBC):
    u = TrialFunction(space)
    v = TestFunction(space)
    x = SpatialCoordinate(space.cell())

    t = initialTime

    f_c_model = None
    if hasattr(Model, "F_c"):
        f_c_model = inner(Model.F_c(t, x, u), grad(v)) # -div F_c v
    if hasattr(Model, "S_e"):
        # there is an issue with S_e returning 'zero' and zero*dx leading to UFL error
        se = (
            inner(as_vector(Model.S_e(t, x, u, grad(u))), v)
        )  # (-div F_c + S_e) * v
        if f_c_model is not None:
            f_c_model += se
        else:
            f_c_model = se

    f_v_model = None
    if hasattr(Model, "F_v"):
        f_v_model = inner(Model.F_v(t, x, u, grad(u)), grad(v)) # -div F_v v

    if hasattr(Model, "S_i"):
        si = inner(as_vector(Model.S_i(t, x, u, grad(u))), v) # (-div F_v + S_i) v
        if f_v_model is not None:
            f_v_model += si
        else:
            f_v_model = si

    # need to extract boundary information from Model
    (
        boundary_flux_cs,
        boundary_flux_vs,
        boundary_values,
        hasBoundaryValue,
    ) = boundaries_ufl(Model, space, t)

    dirichletBCs = [
        DirichletBC(space, item[1], item[0]) for item in boundary_values.items()
    ]
    boundary_flux_vs = -sum(
        [inner(item[1], v) * ds(item[0]) for item in boundary_flux_vs.items()]
    )  # keep all forms on left hand side
    boundary_flux_cs = -sum(
        [inner(item[1], v) * ds(item[0]) for item in boundary_flux_cs.items()]
    )  # keep all forms on left hand side

    if f_c_model:
        f_c_model = f_c_model * dx
    if f_v_model:
        f_v_model = f_v_model * dx
    # !!! fix issue with f_?_model==zero not being a form
    return (
        f_c_model,
        f_v_model,
        {
            "dirichletBCs": dirichletBCs,
            "boundary_flux_cs": boundary_flux_cs,
            "boundary_flux_vs": boundary_flux_vs,
            "hasBoundaryValue": hasBoundaryValue,
        },
    )


def model2ufl(
    Model, space, initialTime=0, DirichletBC=DirichletBC, *, returnFull=False
):
    class M(Model):
        if hasattr(Model, "S_e"):

            def S_e(t, x, U, DU):
                return -Model.S_e(t, x, U, DU)

        if hasattr(Model, "S_i"):

            def S_i(t, x, U, DU):
                return -Model.S_i(t, x, U, DU)

        if hasattr(Model, "F_c"):

            def F_c(t, x, U):
                return -Model.F_c(t, x, U)

    f_c_model, f_v_model, boundary_model = model_ufl(M, space, initialTime, DirichletBC)
    boundary_model["boundary_flux_cs"] = -boundary_model["boundary_flux_cs"]
    form = boundary_model["boundary_flux_cs"] + boundary_model["boundary_flux_vs"]
    if f_c_model is not None:
        form += f_c_model
    if f_v_model is not None:
        form += f_v_model

    if not returnFull:
        return [form == 0, *boundary_model["dirichletBCs"]]
    else:
        boundary_model["f_c_model"] = f_c_model
        boundary_model["f_v_model"] = f_v_model
        boundary_model["form"] = form
        return boundary_model



import numpy as np

try:
    import pygmsh
except ImportError:
    pygmsh = None

try:
    import dune
except ImportError:
    print(
        """
Example code requires dune to run. To install run
pip install dune-fem
    """
    )

from dune.fem.function import gridFunction
from dune.fem.space import lagrange
from dune.grid import cartesianDomain
from dune.ufl import Constant, Space
from ufl import SpatialCoordinate, sqrt

from ddfem import geometry as gm
from ddfem import GridView
from ddfem.geometry.domain_dune import DomainDune


def getDomain(
    initialRefine,
    version,
    adaptLevels=0,
    epsFactor=4.5,
    smoothing=None,
    *args,
    **kwargs,
):

    gm.SDF.smoothing = smoothing
    shiftx, shifty = sqrt(2) * 1e-6, sqrt(3) * 1e-6
    domain_range = [[-1.7 + shiftx, -1.3 + shifty], [1.7 + shiftx, 1.3 + shifty]]
    initial_gridsize = [170 * 2**initialRefine, 130 * 2**initialRefine]
    h = sqrt(
        ((domain_range[1][0] - domain_range[0][0]) / initial_gridsize[0]) ** 2
        + ((domain_range[1][1] - domain_range[0][1]) / initial_gridsize[1]) ** 2
    )

    def get_eps(h):
        return Constant(epsFactor * h * 0.5 ** (adaptLevels / 2), "epsilon")

    balls = [
        [1, [0, 0], "Center"],
        [0.5, [0, 0.8], "TopCut"],
        [0.5, [0, -0.8], "BotCut"],
        [0.5, [1, 0], "RightAdd"],
        [0.5, [-1, 0], "LeftAdd"],
        [1.4, [0, 0], "Cutoff"],
    ]

    b = [gm.Ball(c[0], c[1], name=c[2]) for c in balls]
    # omega = (b[0] - b[1] - b[2] | b[3] | b[4]) & b[5]

    bulk = b[0] - b[1] - b[2] | b[3] | b[4]
    bulk.name = "bulk"

    left_cutoff = b[5] | gm.Plane([-1, 0], -0.4)
    left_cutoff.name = "left"
    right_cutoff = b[5] | gm.Plane([1, 0], -0.4)
    right_cutoff.name = "right"

    cutoff = left_cutoff & right_cutoff  # b[5]

    omega = bulk & cutoff
    omega.name = "full"

    h_max = h * 3
    h_min = h / 2
    radius = 5

    x = SpatialCoordinate(Space(2))
    sdf = omega(x)

    def spacing(x, y, epsilon):
        r_min = epsilon.value
        r_max = radius * epsilon.value
        dist = np.abs(sdf((x, y)))
        if dist <= r_min:
            return geom.characteristic_length_min
        elif dist >= r_max:
            return geom.characteristic_length_max
        else:
            # Linear
            m = (geom.characteristic_length_max - geom.characteristic_length_min) / (
                r_max - r_min
            )
            return m * (dist - r_min) + geom.characteristic_length_min

    omega.epsilon = get_eps(h)

    if version == "cartesian":
        domain = cartesianDomain(*domain_range, initial_gridsize)

    elif version == "fitted":
        if pygmsh is None:
            raise AttributeError("'fitted' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h_max
            geom.characteristic_length_min = h_min

            disks = [geom.add_disk([c[1][0], c[1][1], 0.0], c[0]) for c in balls]

            disk_removed1 = geom.boolean_difference(disks[0], disks[1])
            disk_removed2 = geom.boolean_difference(disk_removed1, disks[2])

            disk_add = geom.boolean_union([disk_removed2, disks[3], disks[4]])
            shape = geom.boolean_intersection([disk_add, disks[5]])

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y, omega.epsilon),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }

    elif version == "dune_adaptive":
        gridsize = [int(j * h / h_max) for j in initial_gridsize]
        domain = cartesianDomain(*domain_range, gridsize)

    elif version == "gmsh_adaptive":
        # in this version we need to adapt the epsilon since using the
        # coarse grid h is not really useful
        if pygmsh is None:
            raise AttributeError("'gmsh_adaptive' requires install pygmsh")
        with pygmsh.occ.Geometry() as geom:
            geom.characteristic_length_max = h_max
            geom.characteristic_length_min = h_min

            geom.add_rectangle(
                [domain_range[0][0], domain_range[0][1], 0.0],
                domain_range[1][0] - domain_range[0][0],
                domain_range[1][1] - domain_range[0][1],
            )

            geom.set_mesh_size_callback(
                lambda dim, tag, x, y, z, lc: spacing(x, y, omega.epsilon),
                ignore_other_mesh_sizes=True,
            )
            mesh = geom.generate_mesh()
            points, cells = mesh.points, mesh.cells_dict
            domain = {
                "vertices": points[:, :2].astype(float),
                "simplices": cells["triangle"].astype(int),
            }
    else:
        raise ValueError("invalid mesh type")

    if not version == "dune_adaptive":
        gridView = GridView(domain)
    else:
        refinements = int(2 * np.log2(h_max / h_min))
        gridView = GridView(domain,omega,
                      factor=radius,
                      refinements=refinements,
                      adaptLevels=adaptLevels)
        h_min = h_max * 0.5 ** (refinements / 2)
        omega.epsilon.value = get_eps(h_min)

    domain = DomainDune(omega, gridView)

    print(
        f"h_max={h_max}, h_min={h_min * 0.5 ** (adaptLevels / 2)}, epsilon={omega.epsilon.value}"
    )

    return gridView, domain



from .arc import Arc
from .ball import Ball
from .box import Box
from .pie import Pie
from .plane import Plane
from .primitive_base import (
    SDF,
    Extrusion,
    Intersection,
    Invert,
    Revolution,
    Rotate,
    Round,
    Scale,
    Subtraction,
    Translate,
    Union,
    Xor,
)
from .vesica import Vesica



from dune.fem.function import gridFunction
from matplotlib import pyplot as plt
from ufl import SpatialCoordinate, triangle


def plotSolution(domain, uh, figsize, **kwargs):
    x = SpatialCoordinate(triangle)
    fig, axs = plt.subplots(1, 2, figsize=figsize)
    uh.plot(figure=(fig, axs[0]), **kwargs)
    gridFunction(uh * domain.chi(x)).plot(figure=(fig, axs[1]), **kwargs)
    for a in axs:
        gridFunction(domain.phi(x), gridView=uh.space.gridView).plot(
            onlyContours=True,
            contours=[0.5],
            gridLines=None,
            contourWidth=2,
            contourColor="black",
            figure=(fig, a),
        )


def plotSdfTree(SDF, gridView, figsize, maxDepth=None, **kwargs):
    x = SpatialCoordinate(triangle)
    fig, axs = plt.subplots(1, 2, figsize=figsize)

    gridFunction(SDF(x), gridView=gridView).plot(figure=(fig, axs[0]), **kwargs)

    def plot_recursive(childSDF, depth):
        if maxDepth is not None and depth > maxDepth:
            return

        gridFunction(childSDF(x), gridView=gridView).plot(
            onlyContours=True,
            contours=[0],
            gridLines=None,
            contourWidth=2,
            contourColor="black",
            figure=(fig, axs[0]),
        )

        for child in childSDF.child_sdf:
            plot_recursive(child, depth + 1)

    plot_recursive(SDF, 0)
    gridFunction(SDF(x), gridView=gridView).plot(
        onlyContours=True,
        contours=[0],
        gridLines=None,
        contourWidth=2,
        contourColor="white",
        figure=(fig, axs[0]),
    )
    gridFunction(SDF.phi(x), gridView=gridView).plot(
        figure=(fig, axs[1]), cmap="coolwarm", **kwargs
    )


def plotBndExt(domain, Model, uh, figsize, t=0, **kwargs):
    from .boundary import BoundaryTerms

    x = SpatialCoordinate(triangle)

    bt = BoundaryTerms(Model, domain)
    bndExt = bt.BndValueExt(t, x, uh)

    fig, axs = plt.subplots(1, figsize=figsize)
    gridFunction(bndExt, gridView=uh.space.gridView).plot(figure=(fig, axs), **kwargs)

    gridFunction(domain(x), gridView=uh.space.gridView).plot(
        onlyContours=True,
        contours=[0],
        gridLines=None,
        contourWidth=2,
        contourColor="black",
        figure=(fig, axs),
    )



from ufl import as_vector, dot

from .primitive_base import SDF


class Plane(SDF):
    def __init__(self, normal, offset, *args, **kwargs):
        if isinstance(normal, (list, tuple)):
            normal = as_vector(normal)
        self.normal = normal

        self.offset = offset

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Plane({self.normal}, {self.offset}, {self._repr_core()})"

    def sdf(self, x):
        return dot(x, self.normal) + self.offset



import sys, os, urllib
try:
    import dune
except ImportError:
    dune = None
import ddfem

from importlib.metadata import version

def getVersion():
    femVersion = version("dune.fem")
    while True:
        dataPath = os.path.join(ddfem.__path__[0],"data",femVersion)
        if os.path.exists(dataPath):
            return femVersion
        ver = femVersion.split(".")
        if "dev" in ver[-1]:
            ver[-1] = 100 # try to find the largest matching release version
        lower = int(ver[-1])-1
        if lower < 0:
            raise RuntimeError(f"no suitable data for {version("dune.fem")} found")
        ver[-1] = str(lower)
        femVersion = ".".join(ver)

def main():
    if dune:
        from dune.fem.utility import FemThreadPoolExecutor
        from dune.generator import builder
    
        module = "intro"
        workers = 4
        if len(sys.argv)>1:
            workers = int(sys.argv[1])
            if len(sys.argv)>2:
                module = sys.argv[2]

        builder.initialize()

        femVersion = getVersion()
        dataPath = os.path.join(ddfem.__path__[0],"data",femVersion)
        print(f"obtaining data from {dataPath}")
        onlyCompile = []
        if not module == "all":
            filename = os.path.join(dataPath,module+".modules")
            with open(filename,"r") as f:
                onlyCompile = f.readlines()
            onlyCompile = set([c.strip()+".cc" for c in onlyCompile])
        else:
            onlyCompile = os.listdir(dataPath)
        with FemThreadPoolExecutor(max_workers=workers) as executor:
            for file in onlyCompile:
                filename = os.path.join(dataPath,os.fsdecode(file))
                if ( not filename.endswith(".cc") or
                     not os.path.exists(filename) ):
                    continue
                with open(filename,"r") as f:
                    executor.submit( builder.load, file.replace(".cc",""), f.read(), None )

        # make sure things are safe on Mac
        # https://docs.python.org/3/library/urllib.request.html
        os.environ["no_proxy"] = "*"

        url = "https://gitlab.dune-project.org/dune-fem/ddfem/-/raw/improveDocBuild/doc/XXX?ref_type=heads"
        if module == "all":
            modules = ["boundary","examples","intro",
                       "transformers", "geometry"]
        else:
            modules = [module]
        for filebase in modules:
            a = urllib.request.urlretrieve(
                 url.replace("XXX",f"{filebase}.py"),
                 filename=f"{filebase}.py")
            b = urllib.request.urlretrieve(
                 url.replace("XXX",f"{filebase}_nb.ipynb"),
                 filename=f"{filebase}_nb.ipynb"
                 )

if __name__ == "__main__":
    main()



from ufl import as_vector

from .helpers import ufl_length
from .primitive_base import ORIGIN, SDF


class Ball(SDF):
    def __init__(self, radius, center=ORIGIN, *args, **kwargs):
        self.radius = radius

        if isinstance(center, (list, tuple)):
            center = as_vector(center)
        self.center = center

        super().__init__(*args, **kwargs)

    def __repr__(self):
        return f"Ball({self.radius}, {self.center}, {self._repr_core()})"

    def sdf(self, x):
        # Note ignore z, if center 2d
        xx = as_vector([x[i] for i in range(len(self.center))])
        center_x = xx - self.center
        return ufl_length(center_x) - self.radius
